from flask import Flask, jsonify, request
import os
import re
import shutil
import pandas as pd
from tpot import TPOTClassifier
from sklearn.model_selection import train_test_split
app = Flask(__name__)
params = {}

@app.route('/set', methods=['post'])
def set_params():
    """ Set training parameters """
    data = request.get_json()
    global params
    for key, value in data.items():
        params[key] = value
    if 'metric' not in params or 'dataset' not in params or 'target' not in params:
        raise Exception("data not complete, need to include metric, dataset, and target")

    return jsonify(params), 201

@app.route('/results/<dataset_name>', methods=['GET'])
def get_result(dataset_name: str):
    """ Train the auto-sklearn and return the result in json format """
    global params
    shutil.rmtree('/tmp', ignore_errors=True)
    null_values = ['', ' ']
    data_path = os.path.join("/data/raw", params['dataset'])
    if not os.path.isfile(data_path):
        return jsonify({"result": "file not exist"}, 200)

    file_type = re.search("[^\.]*$", data_path).group()
    if file_type == "csv":
        df = pd.read_csv(data_path, na_values=null_values)
    elif file_type in ["xls,", "xlsx,", "xlsm,", "xlsb,", "odf,", "ods,", "odt"]:
        df = pd.read_excel(data_path, na_values=null_values)
    else:
        return jsonify({"error": "file type of {} not supported".format(dataset_name)}), 400

    headers = list(df.columns.values)
    target = params["target"]
    headers.remove(target)

    X = df[headers]
    y = df[target]

    # The AutoML part
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, stratify=y, test_size=0.1
    )
    tpot = TPOTClassifier(generations=5, population_size=50, verbosity=2, random_state=42)
    # f1_macro, roc_auc
    if params["metric"] == "auroc":
        tpot._setup_scoring_function('roc_auc')
    elif params["metric"] == "macro_f1":
        tpot._setup_scoring_function('f1_macro')
    else:
        raise Exception("Evaluation metric {} is not yet supported".format(params["metric"]))
    tpot.fit(X_train, y_train)

    return jsonify({"result": tpot.score(X_test, y_test)}), 200

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8082, debug=True)


from typing import Dict
from pprint import pprint
import json
# import re
import os
from sys import exit

from LDAword2vec import LDAword2vec

from .filter_utils import read_txt_files_to_corpus, count_top_models, get_query, find_best_pipes, get_accuracy
from ...processing.utils import Result


def get_ranking(datasets, model_types, topn=3):
    """ Get the ranking from model description and dataset description """
    LLM_result = {}
    for dataset in datasets:
        dataset_result = {}
        LLM_result[dataset] = dataset_result
        query = get_query(dataset)
        for strategy in strategies:
            method_result = {}
            for model_type in model_types:
                model_folder_path = os.path.join("../../../..", 'data', 'models', 'dps', strategy, model_type)
                index_folder_path = os.path.join("../../../..", 'data', 'processed', 'dps', strategy, model_type)
                corpus = read_txt_files_to_corpus(model_folder_path)

                ranker = LDAword2vec(index_folder_path, rerun=True)
                model_result = ranker.search(corpus=corpus, query=query, request_doc_num=3)
                method_result[model_type] = model_result
            dataset_result[strategy] = method_result

    return LLM_result

if __name__ == "__main__":
    # methods = ["combined", "sequential", "interactive"]
    # strategies = ["combined", "sequential"]
    strategies = ["combined"]
    # strategies = ["sequential"]

    datasets = ["Pima", "Ljubljana", "Wisconsin", "NHANES"]
    # datasets = ["NHANES"]

    model_types = ["imputer", "resampler", "classifier"]
    metrics = ["auroc", "macro_f1"]
    # metrics = ["auroc"]

    # ground_truth Structure: {dataset: {metric: {model_type: top count}}}
    ground_truth = find_best_pipes(datasets, metrics)

    # formatted_data = json.dumps(ground_truth, indent=4)
    # pprint(formatted_data)

    # LLM_result Structure: {dataset: {strategy: {model_type: model ranking}}}
    LLM_result = get_ranking(datasets, model_types)
    pprint(LLM_result)

    # Initiate LLM_accuracy structure
    LLM_accuracy = {
        strategy:
            {dataset:
                 {metric: {} for metric in metrics}
             for dataset in datasets}
        for strategy in strategies
    }

    for dataset in ground_truth:
        for metric in ground_truth[dataset]:
            for strategy in strategies:
                LLM_accuracy[strategy][dataset][metric] = get_accuracy(ground_truth[dataset][metric], LLM_result[dataset][strategy])

    pprint(LLM_accuracy)



import requests
from .base_model_client import BaseDockerModelClient


class BaseTransformer(BaseDockerModelClient):
    """ Abstract base class for sklearn-like transformers.
    fit          : Ensure the container is running
    transform    : Post the transform request through RESTful API
    fit_transform: Perform both fit and transform
    """

    def payload(self):
        pass

    def transform(self, X, y=None, dockerfile_dir="."):
        """Transform the training data"""
        try:
            self.ensure_container_running()
            payload = {

            }
            response = requests.post(f"{self.api_url}/transform", json=payload)
            response.raise_for_status()
            return response.json()
        finally:
            self.stop_container()

    def fit_transform(self, X, y=None, dockerfile_dir="."):
        self.fit(X, y, dockerfile_dir)
        return self.transform(X, y, dockerfile_dir)




import numpy as np
from src.hybrids.mesa.mesa import Mesa
from src.customclf import clfs
from src.hybrids.mesa.arguments import args
from sklearn.metrics import precision_recall_fscore_support, roc_auc_score, classification_report, \
    average_precision_score
class RunMESA:
    def __init__(self):
        self.mesa = None
        self.metric = None
    def fit(self, X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray, y_test: np.ndarray, clf="ada", metric=None):
        if clf in clfs.keys():
            clf = clfs[clf]
            self.mesa= Mesa(args=args,base_estimator=clf,n_estimators=10)
            self.mesa.meta_fit(X_train, y_train, X_test, y_test)
            self.mesa.fit(X_train, y_train,X_test,y_test)
            self.metric = metric
        else:
            raise "Model {} not defined in model.py".format(clf)
    def predict(self, X_test: np.ndarray = None, y_test: np.ndarray = None):
        if self.metric == "auroc":
            y_proba = self.mesa.predict_proba(X_test)[:, 1]
            auroc = roc_auc_score(y_test, y_proba)
            self.result = auroc
            return auroc
        elif self.metric == "macro_f1":
            y_pred = self.mesa.predict(X_test)
            _, _, f1, _ = (
                precision_recall_fscore_support(y_test, y_pred, average='macro'))
            self.result = f1
            return f1
        else:
            raise ValueError("Metric {} is not supported in {}".format(self.metric, "MESA"))



from .base_model_api import BaseModelAPI
from abc import abstractmethod


class BaseEstimatorAPI(BaseModelAPI):
    """Abstract base class for sklearn-like estimators/classifiers."""

    @abstractmethod
    def fit(self, args, X_train, y_train, X_test, y_test):
        pass

    @abstractmethod
    def predict(self, X):
        pass



import os
import pickle
import numpy as np
import pandas as pd
import logging

from ..processing.utils import DataLoader, Samplar, Result
from ..processing.preprocessing import DataPreprocess
from ..pipelines.customrsp import CustomResamplar
from ..pipelines.customimputation import CustomImputer
from ..pipelines.customclf import CustomClassifier
from ..pipelines.customhbd import CustomHybrid
from ..pipelines.customautoml import CustomAutoML


def average(lst):
    return sum(lst) / len(lst)

class RunPipe:
    """ Run different pipelines and save the trained results
    Parameters
    ----------
    args : The command line arguments that define how the code should run

    Attributes
    ----------
    preprocessor : Split data into features (X) and target (y)
    args : The command line arguments that define how the code should run
    X : Features
    y : Target
    saver : The class to save trained results and load saved results

    """
    def __init__(self, args=None):
        self.preprocessor = None
        self.args = args
        self.X = None
        self.y = None

        self.dataloader = None
        self.saver = None

    def loadData(self):

        # Load data
        logging.info("Loading Start")
        self.dataloader = DataLoader(self.args.dataset, path=self.args.path)
        data = self.dataloader.train_loader()
        logging.info("Loading Done")

        # Load saved result if it exists
        self.saver = Result(str(self.args.train_ratio), self.args.metric, self.args.dataset, dataloader=self.dataloader)
        self.saver.load_saved_result()

        # Proprocess data
        logging.info("Preprocessing Start")
        self.preprocessor = DataPreprocess(data, self.args)

    def stratifiedSample(self, X, y, train_ratio):
        data = pd.concat([X, y], axis=1)
        new_data = data.groupby('Status', group_keys=False).apply(lambda x: x.sample(frac=train_ratio))
        new_data.sort_index(inplace=True)
        new_data.reset_index(inplace=True, drop=True)
        columns = list(new_data.columns.values)
        columns.remove("Status")
        X = new_data[columns].copy()
        y = new_data["Status"].copy()
        return X, y

    def impute_data(self, imp, train_ratio=1.0):
        # impute_file_name = os.path.join("..", "data", "interim", "imp_" + imp + "_.p")
        impute_file_name = os.path.join(self.dataloader.get_interim_data_folder(), self.args.dataset, "imp_" + imp + ".p")

        if os.path.exists(impute_file_name):
            # Load Saved imputation files
            with open(impute_file_name, "rb") as f:
                X, y = pickle.load(f)
        else:
            # Compute imputation files
            X, y = self.preprocessor.preprocess(self.args)
            imputer = CustomImputer(imp)
            X = imputer.fit_transform(X, y)

            # Save imputation results
            data2save = (X, y)
            if not os.path.isdir(os.path.dirname(impute_file_name)):
                os.makedirs(os.path.dirname(impute_file_name))
            with open(impute_file_name, "wb") as f:
                pickle.dump(data2save, f)

        if train_ratio != 1.0:
            X, y = self.stratifiedSample(X, y, train_ratio)
        return X, y

    def fit_automl(self, pipe, train_ratio=1.0):
        if len(pipe) != 1:
            raise ValueError("Pipeline {} length is not correct, not a AutoML method pipeline".format(pipe))

        # Load saved result if it exists
        # if self.saver.is_in(pipe):
        #     result = self.saver.get(pipe)
        #     return result

        automl = pipe[0]

        X, y = self.preprocessor.preprocess(self.args)

        # fit
        logging.info("\t Training of {} Start".format(pipe))
        trainer = CustomAutoML(self.args, automl)
        trainer.train(X, y)
        logging.info("\t Training of {} Ended".format(pipe))

        # Predict
        logging.info("\t Predicting of {} Start".format(pipe))
        result = trainer.predict()
        logging.info("\t Predicting of {} Ended".format(pipe))

        self.args.repeat += 1

        self.saver.append(pipe, result)
        return result


    def fit_hybrid(self, pipe, train_ratio=1.0):
        if len(pipe) != 2:
            raise ValueError("Pipeline {} length is not correct, not a hybrid method pipeline".format(pipe))

        # Load saved result if it exists
        # if self.saver.is_in(pipe):
        #     result = self.saver.get(pipe)
        #     return result

        # Start training pipeline
        imp, hbd = pipe

        X, y = self.impute_data(imp=imp, train_ratio=train_ratio)
        logging.info("Imputation Done")

        results = []
        train_sampler = Samplar(np.array(X), np.array(y))

        for X_train, y_train, X_test, y_test in train_sampler.apply_kfold(self.args.n_splits):
            logging.info("\t Training in fold {} Start".format(self.args.repeat))
            trainer = CustomHybrid(args=self.args, pipe=pipe)
            if self.args.metric not in trainer.hbd.supported_metrics:
                raise ValueError("Metric {} not yet supported for model {}".format(self.args.metric, hbd))
            trainer.train(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            logging.info("\t Training in fold {} Done".format(self.args.repeat))


            trainer.predict(X_test=X_test, y_test=y_test)

            results.append(trainer.result)

            del trainer

            self.args.repeat += 1

        self.saver.append(pipe, average(results))
        return average(results)


    def fit(self, pipe, train_ratio=1.0):
        if len(pipe) != 3:
            raise ValueError("Pipeline {} length is not correct, not a regular method pipeline")

        # Load saved result if it exists
        # if self.saver.is_in(pipe):
        #     result = self.saver.get(pipe)
        #     return result

        # Run the pipeline
        imp, rsp, clf = pipe

        #
        # Imputation level
        #
        X, y = self.impute_data(imp=imp, train_ratio=train_ratio)

        logging.info("Imputation Done")
        train_sampler = Samplar(np.array(X), np.array(y))

        results = []
        for X_train, y_train, X_test, y_test in train_sampler.apply_kfold(self.args.n_splits):
            logging.info("\t Fold {}".format(self.args.repeat))

            #
            # Resampling level
            #
            resamplar = CustomResamplar(X_train, y_train)
            # params = param_loader()
            # sam_ratio = params[imp][rsp][clf] / 10 if params[imp][rsp][clf] else 1
            # sam_ratio = 1
            # if resamplar.need_resample(sam_ratio):
            if resamplar.need_resample():
                logging.info("\t Re-Sampling Started")
                # X_train, y_train = resamplar.resample(self.args, rsp=rsp, ratio=sam_ratio)
                X_train, y_train = resamplar.resample(self.args, rsp=rsp)
                logging.info("\t Re-Sampling Done")

            #
            # Classification level
            #
            logging.info("\t Training in fold {} Start".format(self.args.repeat))
            trainer = CustomClassifier(self.args)
            trainer.train(X_train, y_train, clf=clf)
            logging.info("\t Training in fold {} Done".format(self.args.repeat))

            # Validation of the result
            trainer.predict(X_test, y_test)

            results.append(trainer.result)

            del trainer

            self.args.repeat += 1

        self.saver.append(pipe, average(results))
        return average(results)


if __name__ == "__main__":
    # import logging
    import warnings

    logging.basicConfig(filename='django_frontend.log', level=logging.DEBUG,
                        format='%(asctime)s:%(levelname)s:%(message)s')
    warnings.filterwarnings("ignore")

    class Args:
        def __init__(self):
            self.train_ratio=1.0
            self.n_splits = 10
            self.repeat = 0
            self.dataset = "nhanes.csv"
            self.metric = "auroc"
            self.target = "Status"
    args = Args()
    run_pipe = RunPipe(args)
    # run_pipe.fit("MIRACLE", "mwmote", "lr")
    # print(run_pipe.fit_hybrid(["imp", "hbd"]))
    # print(run_pipe.fit(["imp", "rsp", "clf"]))
    run_pipe.loadData()
    run_pipe.fit_hybrid(["median", "autosmote"])
    # print(run_pipe.fit_automl(["autosklearn"]))



import pandas as pd

# from ..model_server_core.base_model_client import BaseDockerModelClient
from AutoImblearn.components.model_server_core.base_transformer import BaseTransformer
import os


class RunHyperImpute(BaseTransformer):
    # TODO make model parameter work

    def __init__(self, model="median"):
        super().__init__(
            image_name=f"hyperimpute-api",
            container_name=f"{model}_container",
            container_port=8080,
            volume_mounts={
                os.path.join(os.path.dirname(os.path.abspath(__file__)), model):
                    "/code/AutoImblearn/Docker",
            },  # mount current dir
            dockerfile_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "hyperimputebased"),
        )

    @property
    def payload(self):
        return {
            "metric": self.args.metric,
            "model": self.args.model,
        }




import os
from sys import exit

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.impute import KNNImputer
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import MinMaxScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import KBinsDiscretizer

from hyperimpute.plugins.imputers import Imputers

import pandas as pd
import numpy as np
from scipy.stats import norm, gaussian_kde
import matplotlib.pyplot as plt
from ..processing.utils import find_categorical_columns

from AutoImblearn.components.imputers import RunHyperImpute, RunSklearnImpute

# imps = ["MIRACLE"]
# imps = ["median", "knn", "ii", "gain", "MIRACLE", "MIWAE"]
imps = {
    # "dropna": RunSklearnImpute(model="dropna"),
    "median": RunSklearnImpute(model="median"),
    "mean": RunSklearnImpute(model="mean"),
    "knn": RunSklearnImpute(model="knn"),
    "ii": RunSklearnImpute(model="ii"),
    "gain": RunHyperImpute(model="gain"),
    "MIRACLE": RunHyperImpute(model="MIRACLE"),
    "MIWAE": RunHyperImpute(model="MIWAE"),
}

class CustomImputer(BaseEstimator, TransformerMixin):
    def __init__(self, method="median", aggregation=None):
        self.method = method

        if self.method in imps.keys():
            self.imp = imps[self.method]
            if not hasattr(self.imp, "data_folder_path"):
                raise Exception("Model {} does not have data_folder_path attribute".format(self.hbd))
            setattr(self.hbd, "data_folder_path", self.args.path)
        else:
            raise Exception("Model {} not defined in model.py".format(hbd))

        self.aggregation = aggregation

        self.data = None
        self.header_X = None
        self.feature2drop = []
        self.category_columns = None

    def fit(self, data: np.ndarray):
        self.imp.fit(data)

    def transform(self, data: np.ndarray) -> np.ndarray:
        self.imp.transform(data)

    def fit_transform(self, data: np.ndarray) -> np.ndarray:
        self.fit(data)
        data = self.transform(data)
        return data



class _CustomImputer(BaseEstimator, TransformerMixin):
    def __init__(self, method="median", aggregation=None):
        self.method = method
        self.aggregation = aggregation

        self.data = None
        self.header_X = None
        self.feature2drop = []
        self.category_columns = None

    def apply_rounding(self):
        # Add rounding to categorical features after imputation
        self.category_columns = [i for i in self.category_columns if i in self.data.columns.values]
        for column in self.category_columns:
            self.data[column] = self.data[column].round(0)

    def handle_missing(self):
        # Apply imputation to data
        if self.method == "dropna":
            self.data.dropna(inplace=True)

        elif self.method == "median":
            medians = self.data.median()
            self.data = self.data.fillna(medians)

        elif self.method == "mean":
            means = self.data.mean()
            self.data = self.data.fillna(means)

        elif self.method == "knn":
            data_file_name = "knnimputer.csv"
            file_path = os.path.join("../..", "data", "interim", data_file_name)
            if os.path.isfile(file_path):
                self.data[:] = pd.read_csv(file_path)
            else:
                impute = KNNImputer(weights='distance', n_neighbors=1)
                self.data[:] = impute.fit_transform(self.data)
                self.apply_rounding()
                del impute

        elif self.method == "ii":
            data_file_name = "iiimputer.csv"
            file_path = os.path.join("../..", "data", "interim", data_file_name)
            if os.path.isfile(file_path):
                self.data[:] = pd.read_csv(file_path)
            else:
                impute = IterativeImputer(
                )
                self.data[:] = impute.fit_transform(self.data)
                self.apply_rounding()
                del impute

        elif self.method in ['gain', 'MIRACLE']:
            dict_types = dict(self.data.dtypes)
            old_columns = self.data.columns.values

            impute = Imputers().get(self.method.lower())
            self.data = impute.fit_transform(self.data.astype('float32').copy())
            # Change back to old column names
            rename_dict = dict(map(lambda i, j: (i, j), self.data.columns.values, old_columns))
            self.data.rename(rename_dict, axis=1, inplace=True)

            # Change back to old column dtypes
            self.data = self.data.astype(dict_types)
            self.apply_rounding()
            del impute

        elif self.method in ['MIWAE']:
            dict_types = dict(self.data.dtypes)
            old_columns = self.data.columns.values

            impute = Imputers().get(self.method.lower(), random_state=42, batch_size = 128)
            self.data = impute.fit_transform(self.data.astype('float32').copy())
            # Change back to old column names
            rename_dict = dict(map(lambda i, j: (i, j), self.data.columns.values, old_columns))
            self.data.rename(rename_dict, axis=1, inplace=True)

            # Change back to old column dtypes
            self.data = self.data.astype(dict_types)
            self.apply_rounding()
            del impute

        else:
            raise Exception("Error with handling missing value")

    def find_categorical(self, X: pd.DataFrame):
        """ Predict which columns are categorical """
        self.category_columns = find_categorical_columns(X).keys()
        # self.category_columns = []
        # # TODO make this one a feature
        # categorical_threshold = 0.5
        # for feature_name in X.columns.values:
        #     is_categorical = True
        #     # Calculate the derivatives in the Probability Density Function (PDF)
        #     feature = X[feature_name]
        #     unique_values = np.unique(feature)
        #     unique_count = len(unique_values)
        #
        #     # delete empty value from unique_count
        #     if np.isnan(unique_values).any():
        #         unique_count -= 1
        #
        #     # skip binary features
        #     if unique_count == 2:
        #         continue
        #
        #     total_count = feature.shape[0]
        #
        #     # TODO make the second condition a parameter
        #     if unique_count / total_count < 0.05 and unique_count < 20:
        #         self.category_columns.append(feature_name)
        #         continue
        #
        #     # Find unique value count
        #     # feature_min = feature.min()
        #     # feature_max = feature.max()
        #     # x_values = np.linspace(feature_min, feature_max, 1000)
        #     # kde_values = sum(norm(xi).pdf(x_values) for xi in feature)
        #     kde = gaussian_kde(feature)
        #
        #     x_values = np.linspace(feature.min(), feature.max(), 1000)
        #     kde_values = kde(x_values)
        #
        #     # df data structure: [feature value range, value kde_values, derivative]
        #     df = pd.DataFrame(np.transpose(np.vstack((x_values, kde_values))), columns=['X', 'y'])
        #     df = df.assign(derivative=df.diff().eval('y/X'))
        #
        #     max_density = df['y'].max()
        #
        #     # Find where the local minimal is in the PDF
        #     trough_points = []  # trough_points stores the 'major' trough points of PDF
        #     crest_points = []   # crest_points stores the all crest points of PDF
        #
        #     derivative = df['derivative']
        #     for i in range(len(derivative) - 1):
        #         previous_state = derivative[i]
        #         current_state = derivative[i+1]
        #         if current_state * previous_state <= 0 and previous_state < 0:
        #             # if df['y'].iat[i] - feature_min < categorical_threshold * (feature_max - feature_min):
        #             if max_density - df['y'].iat[i] > categorical_threshold * max_density:
        #                 trough_points.append(i)
        #         elif current_state * previous_state <= 0 and previous_state > 0:
        #             crest_points.append(i)
        #
        #     if not trough_points or not is_categorical:
        #         continue
        #
        #     # shrink the size of the major trough points
        #     index = 0
        #     back = None
        #     front = 0
        #     edges = []
        #     for i in range(len(trough_points) - 1):
        #         trough = df['y'].iat[trough_points[i]]
        #         # find the largest crest before point 'i'
        #         end = trough_points[i]
        #         if back is None:
        #             back = 0
        #             while crest_points[index] < end:
        #                 tmp = df['y'].iat[crest_points[index]]
        #                 if tmp > back:
        #                     back = tmp
        #                 index += 1
        #
        #         # find the largest crest after point 'i'
        #         if i == len(trough_points)-1:
        #             end = len(x_values)
        #         else:
        #             end = trough_points[i + 1]
        #
        #         while index < len(crest_points) and crest_points[index] < end:
        #             tmp = df['y'].iat[crest_points[index]]
        #             if tmp > front:
        #                 front = tmp
        #             index += 1
        #
        #         # TODO determine if treat the 0.3 threshold as parameter
        #         if abs(front - trough) > 0.3 * max_density or abs(back - trough) > 0.3 * max_density:
        #             edges.append(df['X'].iat[trough_points[i]])
        #
        #         back = front
        #         front = 0
        #
        #     if len(edges) < 1:
        #         continue
        #
        #     use_KBinsDiscretizer = True
        #     if use_KBinsDiscretizer:
        #         discretizer = KBinsDiscretizer(n_bins=len(edges) + 1, encode='ordinal', strategy='kmeans')
        #         enc = discretizer.fit_transform(feature.to_numpy().reshape(-1, 1)).reshape(1, -1)[0]
        #         X[feature_name] = pd.Series(enc)
        #     else:
        #         def convert2cal(x, edges):
        #             result = 0
        #             while result < len(edges):
        #                 if x <= edges[result]:
        #                     break
        #                 result += 1
        #             return result
        #
        #         X[feature_name] = X[feature_name].map(lambda x: convert2cal(x, edges))
        #
        #     self.category_columns.append(feature_name)
        #
        #     ### Plot the Probability Density Function (PDF)
        #     # plt.fill_between(x_values, kde_values, alpha=0.5)
        #     # plt.plot(feature, np.full_like(feature, -0.1), '|k', markeredgewidth=1)
        #     # plt.show()
        # return self.category_columns

    def transform_categorical(self, columns=None):
        if columns is not None:
            self.category_columns = columns

        # Transform the categorical
        self.category_columns = [i for i in self.category_columns if i in self.data.columns.values]

        self.data = pd.get_dummies(data=self.data, columns=self.category_columns)

    # def data_scaling(self):
    #     """ Scale Features """
    #     min_max_columns = [
    #     min_max_columns = [item for item in min_max_columns if item in self.data.columns.values]
    #
    #     self.data[min_max_columns] = MinMaxScaler().fit_transform(self.data[min_max_columns])
    def data_scaling(self):
        """ Scale Features """
        min_max_columns = [item for item in self.data.columns.values if item not in self.category_columns]

        self.data[min_max_columns] = MinMaxScaler().fit_transform(self.data[min_max_columns])

    def data_aggregation(self):
        # Transform pandas columns with only 2 unique string values to 0 and 1
        for column in self.data.columns:
            unique_values = pd.Series(self.data[column].unique()).dropna()

            # Check if the column has exactly two unique string values
            if len(unique_values) == 2 and self.data[column].dtype == 'object':
                # Map the two unique values to 0 and 1
                mapping = {unique_values[0]: 0, unique_values[1]: 1}
                self.data[column] = self.data[column].map(mapping)

    def fit(self, X, y = None):
        self.data = X
        return self

    def transform(self, X):
        # Predict which columns are categorical
        self.find_categorical(X)

        self.data_aggregation()

        self.transform_categorical()

        self.handle_missing()

        # Scale data to the same interval
        self.data_scaling()

        self.header_X = self.data.columns.values

        return self.data



# print(__package__)

# if __name__ == "__main__":
#     import os
#     import sys
#     current_path = os.path.dirname(os.path.abspath(__file__))
#     parent_path = os.path.dirname(os.path.dirname(current_path))  # Adjust based on actual layout
#     sys.path.insert(0, parent_path)
#
#     # Optionally set package name for relative imports to work
#     __package__ = "AutoImblearn.api"

import logging

from autosmote.classifiers import get_clf, name2f
from autosmote.rl.training import train

from AutoImblearn.components.api import BaseEstimatorAPI

import numpy as np


class RunAutosmoteAPI(BaseEstimatorAPI):
    def get_hyperparameter_search_space(self):
        return {
            "seed": {
                "type": "int",
                "min": 0,
                "max": 10000,
                "default": 1
            },
            "device": {
                "type": "categorical",
                "choices": ["cpu", "cuda:0", "cuda:1"],
                "default": "cuda:0"
            },
            "cuda": {
                "type": "categorical",
                "choices": ["0", "1"],
                "default": "0"
            },
            "xpid": {
                "type": "categorical",
                "choices": ["AutoSMOTE"],
                "default": "AutoSMOTE"
            },
            "undersample_ratio": {
                "type": "int",
                "min": 10,
                "max": 300,
                "default": 100
            },
            "num_instance_specific_actions": {
                "type": "int",
                "min": 1,
                "max": 50,
                "default": 10
            },
            "num_max_neighbors": {
                "type": "int",
                "min": 5,
                "max": 100,
                "default": 30
            },
            "cross_instance_scale": {
                "type": "int",
                "min": 1,
                "max": 10,
                "default": 4
            },
            "num_actors": {
                "type": "int",
                "min": 1,
                "max": 100,
                "default": 40
            },
            "total_steps": {
                "type": "int",
                "min": 100,
                "max": 50000,
                "default": 1000
            },
            "batch_size": {
                "type": "int",
                "min": 2,
                "max": 64,
                "default": 8
            },
            "cross_instance_unroll_length": {
                "type": "int",
                "min": 1,
                "max": 20,
                "default": 2
            },
            "instance_specific_unroll_length": {
                "type": "int",
                "min": 50,
                "max": 1000,
                "default": 300
            },
            "low_level_unroll_length": {
                "type": "int",
                "min": 50,
                "max": 1000,
                "default": 300
            },
            "num_buffers": {
                "type": "int",
                "min": 5,
                "max": 100,
                "default": 20
            },
            "entropy_cost": {
                "type": "float",
                "min": 0.0001,
                "max": 0.01,
                "default": 0.0006
            },
            "baseline_cost": {
                "type": "float",
                "min": 0.0,
                "max": 1.0,
                "default": 0.5
            },
            "discounting": {
                "type": "float",
                "min": 0.9,
                "max": 1.0,
                "default": 1.0
            },
            "learning_rate": {
                "type": "float",
                "min": 1e-5,
                "max": 0.1,
                "default": 0.005,
                "log_scale": True
            },
            "grad_norm_clipping": {
                "type": "float",
                "min": 1.0,
                "max": 100.0,
                "default": 40.0
            },
            "clf": {
                "type": "categorical",
                "choices": list(name2f.keys()),
                "default": "svm"
            },
        }


    def fit(self, args, X_train, y_train, X_test, y_test):
        params = self.dict_to_namespace()
        params.metric = args.metric

        size = X_train.shape[0]
        indices = np.arange(size)
        np.random.shuffle(indices)

        val_idx = indices[:int(size * args.val_ratio)]
        train_idx = indices[int(size * args.val_ratio):]

        train_X, val_X = X_train[train_idx], X_train[val_idx]
        train_y, val_y = y_train[train_idx], y_train[val_idx]

        logging.info("finished parameter setting")

        params.ratio_map = [0.0, 0.25, 0.5, 0.75, 1.0]
        clf = get_clf(params.clf)
        return train(params, train_X, train_y, val_X, val_y, X_test, y_test, clf)

    def predict(self, X):
        return self.result


import os
import re
import pickle
import json
from typing import List, Literal
from pydantic import BaseModel, Field
import docker

import pandas as pd
import numpy as np
# from imblearn.datasets import fetch_datasets
from sklearn.model_selection import StratifiedKFold
from sklearn.cluster import KMeans
from sklearn.preprocessing import KBinsDiscretizer
from scipy.stats import norm, gaussian_kde


# The path for the data folder
DATA_VOLUME_PATH = "../../data"

client = docker.from_env()

# Absolute path on the host to be shared
SHARED_HOST_DIR = os.path.abspath('DATA_VOLUME_PATH')  # or any absolute path you want
os.makedirs(SHARED_HOST_DIR, exist_ok=True)    # ensure the directory exists


class ArgsNamespace(BaseModel):
    dataset: str = Field(default="nhanes.csv")
    target: str = Field(default="Status")  # Prediction target

    # Model parameters
    T_model: Literal["SVM", "LSVM", "lr", "rf", "mlp", "s2sl", "s2sLR", "ensemble", "ada", "bst"] = "lr"
    repeat: int = Field(default=0)

    # Pre-Processing
    aggregation: Literal["categorical", "binary"] = "binary"
    missing: Literal['median', 'mean', 'dropna', 'knn', 'ii', 'gain', 'MIRACLE', 'MIWAE'] = "median"

    # K-Fold
    n_splits: int = Field(default=10)

    # Resample related
    infor_method: Literal["normal", "nothing"] = "normal"
    resampling: bool = Field(default=False)
    resample_method: Literal["under", "over", "combined", "herding", "s2sl_mwmote", "MWMOTE", "smote"] = "under"
    samratio: float = Field(default=0.4)

    # Feature Importance
    feature_importance: Literal["NA", "lime", "shap"] = "NA"

    # GridSearchCV
    grid: bool = Field(default=False)

    # Top k feature
    top_k: int = Field(default=-1)

    # Auto-Imblearn related
    train_ratio: float = Field(default=1.0)
    metric: Literal["auroc", "macro_f1"] = "auroc"
    rerun: bool = Field(default=False)
    exhaustive: bool = Field(default=False)
    path: str = Field(default=None)


def find_categorical_columns(X: pd.DataFrame) -> dict:
    """ Predict which columns are categorical """
    category_columns = {}
    # TODO make this one a feature
    categorical_threshold = 0.5             # Set a threshold to determine if a column is categorical
    for feature_name in X.columns.values:
        is_categorical = True
        # Calculate the derivatives in the Probability Density Function (PDF)
        feature = X[feature_name]

        # Check if the column has mixed data types (numeric and string). Remove string cells if it's mixed type
        is_mixed = feature.apply(lambda x: pd.to_numeric(x, errors='coerce')).notna().any() and \
                   feature.apply(lambda x: isinstance(x, str)).any()
        if is_mixed:
            feature = pd.to_numeric(feature, errors='coerce')   # Convert mixed type column to numeric, coercing errors to NaN

        unique_values = np.unique(feature)
        unique_count = len(unique_values)

        # Delete empty value from unique_count
        if np.isnan(unique_values).any():
            unique_count -= 1

        # skip binary features
        if unique_count == 2:
            continue

        total_count = feature.shape[0]          # Get the total number of values in the column

        # TODO make the second condition a parameter
        # When there is limited amount of values in the column
        if unique_count / total_count < 0.05 and unique_count < 20:
            unique_values.sort()
            edges = [(unique_values[i] + unique_values[i + 1]) / 2 for i in range(unique_count - 1)]        # Calculate the edges between unique values
            category_columns[feature_name] = edges
            continue

        # Find unique value count
        feature = feature[~np.isnan(feature)]       # Remove NaN values from the feature
        kde = gaussian_kde(feature)                 # Generate a Gaussian Kernel Density Estimate (KDE) for the feature

        # Calculate the derivative of the KDE values for feature
        x_values = np.linspace(feature.min(), feature.max(), 1000)
        kde_values = kde(x_values)

        df = pd.DataFrame(np.transpose(np.vstack((x_values, kde_values))), columns=['X', 'y'])
        df = df.assign(derivative=df.diff().eval('y/X'))

        max_density = df['y'].max()

        # Find where the local minimal is in the PDF
        trough_points = []  # trough_points stores the 'major' trough points of PDF
        crest_points = []  # crest_points stores the all crest points of PDF

        derivative = df['derivative']
        for i in range(len(derivative) - 1):
            previous_state = derivative[i]
            current_state = derivative[i + 1]
            if current_state * previous_state <= 0 and previous_state < 0:
                # if df['y'].iat[i] - feature_min < categorical_threshold * (feature_max - feature_min):
                if max_density - df['y'].iat[i] > categorical_threshold * max_density:
                    trough_points.append(i)
            elif current_state * previous_state <= 0 and previous_state > 0:
                crest_points.append(i)

        if not trough_points or not is_categorical:
            continue

        # shrink the size of the major trough points
        index = 0
        back = None
        front = 0
        edges = []
        for i in range(len(trough_points) - 1):
            trough = df['y'].iat[trough_points[i]]
            # find the largest crest before point 'i'
            end = trough_points[i]
            if back is None:
                back = 0
                while crest_points[index] < end:
                    tmp = df['y'].iat[crest_points[index]]
                    if tmp > back:
                        back = tmp
                    index += 1

            # find the largest crest after point 'i'
            if i == len(trough_points) - 1:
                end = len(x_values)
            else:
                end = trough_points[i + 1]

            while index < len(crest_points) and crest_points[index] < end:
                tmp = df['y'].iat[crest_points[index]]
                if tmp > front:
                    front = tmp
                index += 1

            # TODO determine if treat the 0.3 threshold as parameter
            if abs(front - trough) > 0.3 * max_density or abs(back - trough) > 0.3 * max_density:
                edges.append(df['X'].iat[trough_points[i]])

            back = front
            front = 0

        edges.sort()

        if len(edges) < 1:
            continue

        # TODO make this one a selection feature
        # TODO don't change X[feature_name]
        use_KBinsDiscretizer = False
        if use_KBinsDiscretizer:
            discretizer = KBinsDiscretizer(n_bins=len(edges) + 1, encode='ordinal', strategy='kmeans')
            enc = discretizer.fit_transform(feature.to_numpy().reshape(-1, 1)).reshape(1, -1)[0]
            X[feature_name] = pd.Series(enc)
        else:
            def convert2cal(x, edges):
                result = 0
                while result < len(edges):
                    if x <= edges[result]:
                        break
                    result += 1
                return result

            X[feature_name] = X[feature_name].map(lambda x: convert2cal(x, edges))

        # calcuate the edges of the category
        if use_KBinsDiscretizer:
            edges = discretizer.bin_edges_[0].tolist()[1:-1]
            category_columns[feature_name] = edges
        else:
            category_columns[feature_name] = edges

        # print(edges)
        # from sys import exit
        # exit()
        ### Plot the Probability Density Function (PDF)
        # plt.fill_between(x_values, kde_values, alpha=0.5)
        # plt.plot(feature, np.full_like(feature, -0.1), '|k', markeredgewidth=1)
        # plt.show()
    return category_columns


class DataLoader:
    """ DataLoader object to load train, valid and test data from dataset.
        Args:
            dataset (str): Name os the dataset
    """

    def __init__(self,
                 dataset: str='nhanes.csv', # The dataset name for training
                 is_notebook=False,         # If running the code in jupyter notebook
                 path=None,                 # The folder path that contains the dataset
                 ) -> None:
        # self.item_set = set()
        if path is None:
            dir_path = os.path.dirname(os.path.realpath(__file__))
            self.data_folder = os.path.join(dir_path, "../..", 'data')
            self.path = os.path.join(self.data_folder, 'raw', dataset)
        else:
            self.data_folder = path
            self.path = os.path.join(self.data_folder, 'raw', dataset)

        self.name = os.path.basename(self.path)


        if is_notebook:
            self.path = os.path.join("../../../..", self.path)
        self.header = []


    def get_data_folder(self):
        return self.data_folder

    def get_interim_data_folder(self):
        return os.path.join(self.get_data_folder(), 'interim')

    def get_processed_data_folder(self):
        return os.path.join(self.get_data_folder(), 'processed')

    def get_raw_data_folder(self):
        return os.path.join(self.get_data_folder(), 'raw')

    def get_models_data_folder(self):
        return os.path.join(self.get_data_folder(), 'models')

    def get_models_dp_folder(self):
        return os.path.join(self.get_models_data_folder(), 'dps')

    def get_models_dp_origins_folder(self):
        return os.path.join(self.get_models_data_folder(), 'origins')

    def get_dataset_dp_folder(self):
        return os.path.join(self.data_folder, 'datasets')

    def get_data_description(self, data_name):
        dp_folder = self.get_dataset_dp_folder()
        dp_path = os.path.join(dp_folder, 'dps.json')

        with open(dp_path) as f:
            dps = json.load(f)
        return dps[data_name]

    def train_loader(self, has_header=True) -> pd.DataFrame:
        """ Load training data
        :returns:
            df: whole data
        """
        if os.path.isfile(self.path):

            null_values = ['', ' ']

            # filetype = re.search("[^\.]*$", self.path).group()
            filetype = re.search(r"[^.]*$", self.path).group()

            if has_header:
                if filetype == "csv":
                    df = pd.read_csv(self.path, na_values=null_values)
                elif filetype == "xlsx":
                    df = pd.read_excel(self.path, na_values=null_values)
            else:
                if filetype == "csv":
                    df = pd.read_csv(self.path, na_values=null_values, header=None)
                elif filetype == "xlsx":
                    df = pd.read_excel(self.path, na_values=null_values, header=None)
            self.header = list(df.columns.values)
            return df
        # datasets = fetch_datasets()
        # if self.name in datasets.keys():
        #     df = datasets[self.name]
        #     # TODO set up dataset to accept target column value
        #     data = pd.DataFrame(data=df.data)
        #     data.columns = data.columns.astype(str)
        #     target = pd.DataFrame(data=df.target)
        #     target = target.rename(columns={0:"Status"})
        #     df = pd.concat([data, target], axis=1).copy()
        #     return df
        else:
            raise ValueError("There is no corresponding dataset of: {}".format(self.name))


class Samplar:
    """ Samplar oject to split data
        Args:
            X (np.ndarray):
            Y (np.ndarray):
    """
    def __init__(self, X: np.ndarray, Y: np.ndarray):

        self.X = X
        self.Y = Y


    def apply_kfold(self, split_num):
        """Apply stratified cross validation to dataset"""
        skf = StratifiedKFold(n_splits=split_num, shuffle=True, random_state=42)
        for train_index, test_index in skf.split(self.X, self.Y):
            X_train, X_test = self.X[train_index], self.X[test_index]
            Y_train, Y_test = self.Y[train_index], self.Y[test_index]
            yield (X_train, Y_train, X_test, Y_test)


def param_loader():
    # Load resampling strategy find manually
    param_file = os.path.join("../..", 'data', 'interim', "params.p")
    with open(param_file, "rb") as f:
        params = pickle.load(f)
    return params


class Result:
    """ The result saving class
    Parameters
    ----------
    train_ratio : the percentage of data used for training
        See main.py for more information.

    metric : The evaluation metric choosed for the training
        Mainly choose from "auroc" and "macro_f1"

    dataset : The name of the datased used for training

    Attributes
    ----------
    saved_file_path : The file path to save the results

    saved_result : The trained results
    """

    def __init__(self, train_ratio, metric, dataset, dataloader: DataLoader = None):
        saved_file_name = "{}_saved_pipe_{}_{}.json".format(dataset, metric, str(train_ratio))

        if dataloader is None:
            pass
        else:
            self.saved_file_path = os.path.join(dataloader.get_processed_data_folder(), saved_file_name)
        self.saved_result = {}

    def load_saved_result(self):
        # load data from file
        if os.path.isfile(self.saved_file_path):
            with open(self.saved_file_path, "r") as f:
                self.saved_result = json.load(f)
        else:
            self.saved_result = {}

    def is_in(self, pipe):
        # Check if pipe in result
        print(pipe)
        try:
            tmp = self.saved_result[pipe[0]]
            for i in range(1, len(pipe)):
                tmp = tmp[pipe[i]]
            return True
        except KeyError:
            return False

    def append(self, pipe, score):
        # append pipeline into result
        if len(pipe) == 1:
            automl = pipe[0]
            if automl not in self.saved_result:
                self.saved_result[automl] = score
        elif len(pipe) == 2:
            imp, hbd = pipe
            if imp not in self.saved_result:
                self.saved_result[imp] = {}
            if hbd not in self.saved_result[imp]:
                self.saved_result[imp][hbd] = score
        elif len(pipe) == 3:
            imp, rsp, clf = pipe
            if imp not in self.saved_result:
                self.saved_result[imp] = {}
            if rsp not in self.saved_result[imp]:
                self.saved_result[imp][rsp] = {}
            if clf not in self.saved_result[imp][rsp]:
                self.saved_result[imp][rsp][clf] = score
        else:
            raise ValueError("Length of pipe {} is not correct".format(pipe))
        self.save2file()


    def get(self, pipe: List):
        # get the result
        result = self.saved_result[pipe[0]]
        for i in range(1, len(pipe)):
            result = result[pipe[i]]
        return result
        # imp, rsp, clf = pipe
        # return self.saved_result[imp][rsp][clf]

    def save2file(self):
        # save result to json file
        if self.saved_result is None:
            raise ValueError("Please create saved_result first before save.")
        with open(self.saved_file_path, "w") as f:
            json.dump(self.saved_result, f, indent=4)


if __name__ == "__main__":
    # a = param_loader()
    # print(a)

    # a = Result(1.0, "auroc", "test")
    # a.load_saved_result()
    # print(a.saved_result)
    # print(a.is_in(["automl"]))
    # print(a.is_in(["imp", "hbd"]))
    # print(a.is_in(["imp", "rsp", "clf"]))
    # print(a.get(["imp", "rsp", "clf"]))


    dataloader = DataLoader("nhanes.csv")
    data = dataloader.train_loader()
    find_categorical_columns(data)

    args = ArgsNamespace(
        dataset="my_data.csv",
        target="Outcome",
        T_model="rf",
        repeat=10,
        aggregation="categorical",
        missing="mean",
        n_splits=5,
        infor_method="nothing",
        resampling=True,
        resample_method="over",
        samratio=0.6,
        feature_importance="lime",
        grid=True,
        top_k=20,
        train_ratio=0.75,
        metric="macro_f1",
        rerun=True,
        exhaustive=True
    )

    print(args.dataset)



import pandas as pd

# from ..model_server_core.base_model_client import BaseDockerModelClient
from AutoImblearn.components.model_server_core.base_estimator import BaseEstimator
import os


class RunAutoSmote(BaseEstimator):
    def __init__(self):
        super().__init__(
            image_name="autosmote-api",
            container_name="autosmote_container",
            container_port=8080,
            volume_mounts={
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "autosmote"):
                    "/code/AutoImblearn/Docker",
            },  # mount current dir
            dockerfile_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "autosmote"),
        )

    @property
    def payload(self):
        return {
            "metric": self.args.metric,
        }


if __name__ == "__main__":

    input_csv = "pima-indians-diabetes-missing.csv "
    label_col = "Status"

    runner = RunAutoSmote()

    print("[✓] Training model...")
    runner.fit(input_csv, y_train=label_col)

    print("[✓] Predicting (generating balanced data)...")
    result_df = runner.predict(input_csv)

    output_csv = input_csv.replace(".csv", "_balanced.csv")
    result_df.to_csv(output_csv, index=False)
    print(f"[✓] Saved balanced output to: {output_csv}")



from .customimputation import imps as imputers
from .customautoml import automls
from .customhbd import hbds as hybrid_imbalanced_classifiers
from .customclf import clfs as classifiers
from .customrsp import rsps as resamplers



import torch
from torch import nn
from torch.nn import functional as F

class CrossInstanceNet(nn.Module):
    def __init__(self, observation_space, num_actions):
        super().__init__()
        self.state_dim = observation_space["src_features"].shape[0]
        self.state_hidden = 128
        self.num_actions = num_actions

        self.fc_0 = nn.Linear(self.state_dim, self.state_hidden)
        self.fc_1 = nn.Linear(self.state_hidden, self.state_hidden)

        self.policy = nn.Linear(self.state_hidden, num_actions)
        self.baseline = nn.Linear(self.state_hidden, 1)

    def forward(self, inputs, device="cpu"):
        x = inputs["src_features"]
        T, B, *_ = x.shape
        x = torch.flatten(x, 0, 1)  # Merge time and batch.
        x = F.relu(self.fc_0(x))
        x = F.relu(self.fc_1(x))

        policy_logits = self.policy(x)
        baseline = self.baseline(x)

        if self.training:
            #print(torch.argmax(policy_logits, dim=1))
            action = torch.multinomial(F.softmax(policy_logits, dim=1), num_samples=1)
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        return dict(policy_logits=policy_logits, baseline=baseline, action=action)

class InstanceSpecificNet(nn.Module):
    def __init__(self, observation_space, num_actions):
        super().__init__()
        self.state_dim = observation_space["src_features"].shape[0]
        self.state_hidden = 128
        self.num_actions = num_actions

        self.fc_0 = nn.Linear(self.state_dim, self.state_hidden)
        self.fc_1 = nn.Linear(self.state_hidden, self.state_hidden)

        self.policy = nn.Linear(self.state_hidden, num_actions)
        self.baseline = nn.Linear(self.state_hidden, 1)

    def forward(self, inputs, device="cpu"):
        x = inputs["src_features"]
        T, B, *_ = x.shape
        x = torch.flatten(x, 0, 1)  # Merge time and batch.
        x = F.relu(self.fc_0(x))
        x = F.relu(self.fc_1(x))

        policy_logits = self.policy(x)
        baseline = self.baseline(x)

        if self.training:
            #print(torch.argmax(policy_logits, dim=1))
            action = torch.multinomial(F.softmax(policy_logits, dim=1), num_samples=1)
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        return dict(policy_logits=policy_logits, baseline=baseline, action=action)

class LowLevelNet(nn.Module):
    def __init__(self, observation_space, num_ratios, device="cpu"):
        super().__init__()
        self.state_dim = observation_space["src_features"].shape[0]
        self.num_actions, self.action_dim = observation_space["neighbor_features"].shape
        self.num_actions *= num_ratios # The distance is 0, 0.25, 0.5, 0.75, 1.0
        self.hidden = 128
        self.num_ratios = num_ratios

        # Policy
        self.p_fc_0 = nn.Linear(self.state_dim+self.action_dim+num_ratios, self.hidden)
        self.p_fc_1 = nn.Linear(self.hidden, self.hidden)
        self.policy = nn.Linear(self.hidden, 1)

        # Baseline
        self.v_fc_0 = nn.Linear(self.state_dim, self.hidden)
        self.v_fc_1 = nn.Linear(self.hidden, self.hidden)
        self.baseline = nn.Linear(self.hidden, 1)

        self.distance_features = []
        for i in range(self.num_ratios):
            tmp = [0 for _ in range(self.num_ratios)]
            tmp[i] = 1
            self.distance_features.append(tmp)
        self.distance_features = torch.tensor(self.distance_features, dtype=torch.float32)
        self.distance_features = self.distance_features.to(device)
    def forward(self, inputs, considered_neighbors, device="cpu"):
        state_features = inputs["src_features"]
        T, B, *_ = state_features.shape
        state_features = torch.flatten(state_features, 0, 1)  # Merge time and batch.
        neighbor_features = inputs["neighbor_features"]
        neighbor_features = torch.flatten(neighbor_features, 0, 2)  # Merge time, batch, and num_actions.
        neighbor_features = neighbor_features.repeat(1, self.num_ratios).reshape(-1, self.action_dim)
        distance_features = self.distance_features.repeat(neighbor_features.shape[0]//self.num_ratios, 1)
        action_features = torch.cat((neighbor_features, distance_features), -1)
        # Policy
        x = state_features.repeat(1,self.num_actions).view(-1,state_features.shape[1])
        x = torch.cat((x, action_features), -1)
        x = F.relu(self.p_fc_0(x))
        x = F.relu(self.p_fc_1(x))
        policy_logits = self.policy(x)
        policy_logits = policy_logits.view(-1, self.num_actions)

        # Baseline
        baseline = F.relu(self.v_fc_0(state_features))
        baseline = F.relu(self.v_fc_1(baseline))
        baseline = self.baseline(baseline)

        if self.training:
            action = torch.multinomial(F.softmax(policy_logits[:, :considered_neighbors*self.num_ratios], dim=1), num_samples=1)
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        return dict(policy_logits=policy_logits, baseline=baseline, action=action)







import time
import json
import os
import docker
import requests
import pandas as pd
import numpy as np
from pathlib import Path
import logging
import socket
from abc import abstractmethod


def get_free_host_port():
    """Find a free port on the host (works from inside a container using DooD)"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0))  # let OS pick a free port
        return s.getsockname()[1]

def is_port_free(port):
    """Test if the port is free"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex(('localhost', port)) != 0

class BaseDockerModelClient:
    """Build and run the docker container for the selected model"""
    def __init__(self, image_name, container_name, container_port=5000, volume_mounts=None, dockerfile_dir=None):
        self.image_name = image_name
        self.container_name = container_name
        self.container_port = container_port
        self.volume_mounts = volume_mounts or {}
        self.client = docker.from_env()

        # set container url
        self.host_port = get_free_host_port()
        if not is_port_free(self.host_port):
            raise Exception(f"Port {self.host_port} is not free")
        self.api_url = f"http://localhost:{self.host_port}"

        # Dockerfile
        self.dockerfile_dir = dockerfile_dir

        self.args = None

    @property
    @abstractmethod
    def payload(self):
        """Subclasses must return a dict with key arguments."""
        pass

    def is_container_running(self):
        """
        Check if the Docker container with self.container_name exists and is running.
        Returns True if running, False otherwise.
        """
        try:
            container = self.client.containers.get(self.container_name)
            return container.status == 'running'
        except docker.errors.NotFound:
            # Container does not exist
            return False
        except Exception as e:
            # Other error (log it if needed)
            import logging
            logging.warning(f"Error checking container status: {e}")
            return False

    def build_image(self):
        """Build image based on Dockerfile provided"""
        logging.info(f"[⛏️ ] Building image '{self.image_name}' from {self.dockerfile_dir}...")
        # self.client.images.build(path=dockerfile_dir, tag=self.image_name)
        try:
            self.client.images.get(self.image_name)
            logging.info('found prebuilt image')
        except docker.errors.ImageNotFound:
            logging.info("Building AutoSMOTE 1.0 image")
            # TODO update this image build process to automate image genenration
            self.client.images.build(path=self.dockerfile_dir, tag=self.image_name, nocache=True)

        logging.info(f"[✓] Image '{self.image_name}' is available now.")

    def start_container(self):
        """
        Start the container
        """
        logging.info(f"[🚀] Starting container '{self.container_name}'...")
        # binds = {
        #     str(Path(local).resolve()): {'bind': container, 'mode': 'rw'}
        #     for local, container in self.volume_mounts.items()
        # }
        binds = {}
        for local, target in self.volume_mounts.items():
            host_path = str(Path(local).resolve())
            if isinstance(target, str):
                binds[host_path] = {'bind': target, 'mode': 'rw'}
            elif isinstance(target, dict):
                binds[host_path] = target  # already formatted correctly
            else:
                raise TypeError(f"Invalid volume mount target for {host_path}: {target}")

        self.container = self.client.containers.run(
            image=self.image_name,
            name=self.container_name,
            ports={f"{self.container_port}/tcp": self.host_port},
            volumes=binds,
            entrypoint=["python3","-m", "app"],
            working_dir='/code/AutoImblearn/Docker',
            detach=True
        )

    def stop_container(self):
        """
        Stop the running container
        """
        logging.info(f"[🧹] Stopping container '{self.container_name}'...")
        try:
            container = self.client.containers.get(self.container_name)
            container.stop()
            container.remove()
            logging.info(f"[✓] Container '{self.container_name}' removed.")
        except docker.errors.NotFound:
            logging.info("[!] Container not found, skipping.")

    def wait_for_api(self, timeout=60):
        """
        Wait for the container's API to become responsive.
        Raise an error if the container exits or doesn't respond in time.
        """
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                # Check container state
                container = self.client.containers.get(self.container_name)
                container.reload()
                if container.status == "exited":
                    logs = container.logs().decode(errors="ignore")
                    raise RuntimeError(f"Container exited prematurely.\nLogs:\n{logs}")

                # Check API health
                response = requests.get(f"{self.api_url}/health")
                if response.status_code == 200:
                    return True
            except requests.exceptions.ConnectionError:
                pass  # wait and retry
            except docker.errors.NotFound:
                raise RuntimeError("Container was removed unexpectedly.")

            time.sleep(1)

        raise TimeoutError("API did not become available within timeout period.")

    def ensure_container_running(self):
        """
        Ensure the Docker container is running. Build and start if needed.
        """
        try:
            container = self.client.containers.get(self.container_name)
            if container.status == 'running':
                print(f"Container {self.container_name} is already running.")
                ports = container.attrs['NetworkSettings']['Ports']
                internal = f"{self.container_port}/tcp"
                host_info = ports.get(internal)
                if host_info:
                    self.host_port = int(host_info[0]['HostPort'])
                    self.api_url = f"http://localhost:{self.host_port}"
                    print(f"Reusing existing container on port {self.host_port}")
                    return
                else:
                    raise RuntimeError("Could not find port mapping for the running container.")
            else:
                print(f"Container {self.container_name} is not running. Starting...")
                container.start()
        except docker.errors.NotFound:
            print("Container not found. Building and starting a new one...")
            self.build_image()
            print("Building and starting a new container...")
            self.start_container()

        # self.wait_for_api(host="localhost", port=self.host_port, path='/health')
        self.wait_for_api()
        print("API is ready.")

    def save_data_2_volume(self, data_folder_path: str, dataset_name: str, X_train, y_train=None, X_test=None, y_test=None):
        """
        Save data to docker volume
        """
        data_names = [f"X_train_{self.container_name}.csv", f"y_train_{self.container_name}.csv", f"X_test_{self.container_name}.csv",
                      f"y_test_{self.container_name}.csv"]

        for data, name in zip([X_train, y_train, X_test, y_test], data_names):
            if data is None:
                continue
            data_path = os.path.join(data_folder_path, "interim", dataset_name, name)
            np.savetxt(data_path, data, delimiter=",")

        return data_names

    def fit(self, args,  X_train, y_train=None, X_test=None, y_test=None):
        """
        Fit with training data
        """
        print("inside fit")
        self.args = args
        # Save temporary data files to docker volume
        data_names = self.save_data_2_volume(args.path, args.dataset, X_train, y_train, X_test, y_test)
        print("finished data name")

        # Mount data folder to docker container
        # self.volume_mounts[args.path] = '/data'
        # Mount meta class for flask API
        # self.volume_mounts[os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        #                                 "api")] = '/code/AutoImblearn/api'


        try:
            self.ensure_container_running()
            print("after ensuring container running")

            headers = {"Content-Type": "application/json"}
            response = requests.post(f"{self.api_url}/set", json=self.payload, headers=headers)
            print("after set params")
            if response.status_code != 201:
                raise Exception("There is an error in setting parameters")

            response = requests.post(f"{self.api_url}/train", json=self.payload, headers=headers)
            print("after train")
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logging.error(e)
            # self.stop_container()
            raise e



# from .customimputation import imps
# from .customautoml import automls
# from .customhbd import hbds
# from .customclf import clfs
# from .customrsp import rsps
from .main import AutoImblearnTraining


from AutoImblearn.components.api import BaseTransformerAPI
import logging
import numpy as np


class RunAutosmoteAPI(BaseTransformerAPI):
    def get_hyperparameter_search_space(self):
        return {
        }


    def fit(self, args, X_train, y_train, X_test, y_test):
        params = self.dict_to_namespace()
        params.metric = args.metric

        size = X_train.shape[0]
        indices = np.arange(size)
        np.random.shuffle(indices)

        val_idx = indices[:int(size * args.val_ratio)]
        train_idx = indices[int(size * args.val_ratio):]

        train_X, val_X = X_train[train_idx], X_train[val_idx]
        train_y, val_y = y_train[train_idx], y_train[val_idx]

        logging.info("finished parameter setting")

        params.ratio_map = [0.0, 0.25, 0.5, 0.75, 1.0]
        clf = get_clf(params.clf)
        return train(params, train_X, train_y, val_X, val_y, X_test, y_test, clf)

    def transform(self, X):
        return X



import argparse
import logging
import os.path
import warnings

import joblib

from .core.runpipe import RunPipe
from .core.autoimblearn import AutoImblearn
from .processing.utils import ArgsNamespace

logging.basicConfig(filename='django_frontend.log', level=logging.DEBUG,
                    format='%(asctime)s:%(levelname)s:%(message)s')

warnings.filterwarnings("ignore")


class AutoImblearnTraining:
    def __init__(self,
                dataset,      # Dataset name
                target,       # Set the name of the prediction target

                # Model parameters
                T_model,      # Traditional models
                repeat,

                # Pre-Processing
                aggregation,
                missing,      # Handle null values

                # K-Fold
                n_splits,     # Number of split in for K-fold

                # Resample related
                infor_method, # Choose how to handle AUDM
                resampling,
                resample_method,
                samratio,     # target sample ratio

                # Feature Importance
                feature_importance,   # Which model to use

                # GridSearchCV
                grid,         # Use Grid search to find best hyper-parameter

                # top k feature
                top_k,        # The number of features to keep

                # Auto-Imblearn related
                train_ratio,  # Only use certain ratio of dataset
                metric,       # Determine the metric
                rerun,        # Re-run the best pipeline found with 100% data
                exhaustive,   # run exhaustive search instead of AutoImblearn
                path = None,  # the path that stores all kinds of data
                ):

        self.args = ArgsNamespace(
            dataset=dataset,
            target=target,
            T_model=T_model,
            repeat=repeat,
            aggregation=aggregation,
            missing=missing,
            n_splits=n_splits,
            infor_method=infor_method,
            resampling=resampling,
            resample_method=resample_method,
            samratio=samratio,
            feature_importance=feature_importance,
            grid=grid,
            top_k=top_k,
            train_ratio=train_ratio,
            metric=metric,
            rerun=rerun,
            exhaustive=exhaustive,
            path=path,
        )

        self.model_dir = os.path.join(self.args.path, "interim", self.args.dataset, "saved_models")
        os.makedirs(self.model_dir, exist_ok=True)
        self.model_path = os.path.join(self.model_dir, "autoimblearn.pkl")

        # Save the result
        self.result = None   # save the final result from training

    def fit(self):
        logging.info("-------------------------")

        for arg, value in sorted(vars(self.args).items()):
            logging.info("Argument {}: {}".format(arg, value))

        # Load the data
        run_pipe = RunPipe(args=self.args)
        run_pipe.loadData()

        # Run Auto-Imblearn to find best pipeline
        checked = {}
        automl = AutoImblearn(run_pipe, metric=self.args.metric)

        if self.args.exhaustive:
            print("Exhaustive search...")
            automl.exhaustive_search(checked=checked, train_ratio=self.args.train_ratio)

        else:
            print("Finding best pipeline...")
            best_pipe, counter, best_score = automl.find_best(checked=checked, train_ratio=self.args.train_ratio)

            print(f'Final result. Best pipe: {" ".join(list(best_pipe))}, counter: {counter}, best score: {best_score}')
            if self.args.train_ratio != 1.0 and self.args.rerun:
                # Re-run the best pipeline with whole dataset to get the output score
                print("Re-running best pipeline")
                best_score = automl.run_best(best_pipe)

            logging.info(
                f'Final result. Best pipe: {" ".join(list(best_pipe))}, counter: {counter}, best score: {best_score}')
        self.result = {'best_pipeline': best_pipe, 'counter': counter, 'best_score': best_score}

    def predict(self):
        return self.result

    def save_model(self):
        if self.result:
            joblib.dump(self.result, self.model_path)

    def load_model(self):
        if os.path.exists(self.model_path):
            self.result = joblib.load(self.model_path)
        else:
            raise FileNotFoundError("No trained model found")



if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('--dataset', type=str, default="nhanes.csv")
    parser.add_argument('--target', default="Status", type=str)  # Set the name of the prediction target

    #
    # Model parameters
    #
    parser.add_argument('--T_model', default="lr",
                        choices=["SVM", 'LSVM', 'lr', 'rf', 'mlp', 's2sl', 's2sLR', 'ensemble', 'ada',
                                 'bst'])  # Traditional models

    parser.add_argument('--repeat', default=0, type=int)

    #
    # Pre-Processing
    #
    parser.add_argument('--aggregation', default="binary", choices=["categorical", "binary"])

    parser.add_argument('--missing', default='median', choices=['median', 'mean', 'dropna', 'knn', 'ii', 'gain', 'MIRACLE', 'MIWAE'],
                        type=str)  # Handle null values

    # K-Fold
    parser.add_argument('--n_splits', default=10, type=int)  # Number of split in for K-fold

    # Resample related
    parser.add_argument('--infor_method', default='normal', choices=['normal', 'nothing'])  # Choose how to handle AUDM

    parser.add_argument('--resampling', default=False, action="store_true")
    parser.add_argument('--resample_method', default="under",
                        choices=['under', 'over', 'combined', 'herding', 's2sl_mwmote', 'MWMOTE', "smote"])
    parser.add_argument('--samratio', default=0.4, type=float)  # target sample ratio

    # Feature Importance
    parser.add_argument('--feature_importance', default='NA', choices=['NA', 'lime', 'shap'],
                        type=str)  # Which model to use

    # GridSearchCV
    parser.add_argument('--grid', default=False, action="store_true")  # Use Grid search to find best hyper-parameter

    # top k feature
    parser.add_argument('--top_k', default=-1, type=int)  # The number of features to keep

    # Auto-Imblearn related
    parser.add_argument('--train_ratio', default=1.0, type=float)  # Only use certain ratio of dataset
    parser.add_argument('--metric', default='auroc', choices=['auroc', 'macro_f1'], type=str)  # Determine the metric
    # parser.add_argument('--rerun', default=False, action="store_true")  # Re-run the best pipeline found with 100% data
    parser.add_argument('--rerun', default=False, action="store_true")  # Re-run the best pipeline found with 100% data
    parser.add_argument('--exhaustive', default=False, action="store_true") # run exhaustive search instead of AutoImblearn

    args = parser.parse_args()

    logging.info("-------------------------")

    for arg, value in sorted(vars(args).items()):
        logging.info("Argument {}: {}".format(arg, value))

    # Load the data
    run_pipe = RunPipe(args=args)
    run_pipe.loadData()

    # Run Auto-Imblearn to find best pipeline
    checked = {}
    automl = AutoImblearn(run_pipe, metric=args.metric)

    if args.exhaustive:
        print("exhaustive search")
        automl.exhaustive_search(checked=checked, train_ratio=args.train_ratio)

    else:
        best_pipe, counter, best_score = automl.find_best(checked=checked, train_ratio=args.train_ratio)

        print("Final result:", best_pipe, args.metric, counter, end=" ")
        if args.train_ratio != 1.0 and args.rerun:
            # Re-run the best pipeline with whole dataset to get the output score
            print("Re-running best pipeline")
            best_score = automl.run_best(best_pipe)

        print(best_score)
        best_pipe = list(best_pipe)
        logging.info("Final result. Best pipe: {}, {}, {}, counter: {}, best score: {}".format(best_pipe[0], best_pipe[1],
                                                                                               best_pipe[2], counter,
                                                                                               best_score))



import os
import json
import requests
from time import sleep
import warnings

import docker
import numpy as np
import pandas as pd
from sklearn.exceptions import ConvergenceWarning

warnings.filterwarnings("ignore", category=ConvergenceWarning)

import logging


class Arguments:
    def __init__(self):
        self.dataset = "imp_knn_.p"
        self.seed = 1
        self.clf = "ada"
        self.metric = "auroc"

        self.device = "cpu"
        self.cuda = "0"

        self.xpid = "AutoSMOTE"
        self.undersample_ratio = 100

        self.num_instance_specific_actions = 10
        self.num_max_neighbors = 30
        self.cross_instance_scale = 4
        self.savedir = "logs"
        self.num_actors = 10
        self.total_steps = 500
        self.batch_size = 8
        self.cross_instance_unroll_length = 2
        self.instance_specific_unroll_length = 300
        self.low_level_unroll_length = 300
        self.num_buffers = 20

        # Loss settings.
        self.entropy_cost = 0.0006
        self.baseline_cost = 0.5
        self.discounting = 1.0

        # Optimizer settings.
        self.learning_rate = 0.005
        # self.learning_rate = 0.0005
        self.grad_norm_clipping = 40.0
        self.val_ratio = 0.2
        self.test_raito = 0.2
        self.data_names = None


class RunAutoSmote:
    def __init__(self):
        self.data_folder_path = None
        self.flags = Arguments()
        os.environ["CUDA_VISIBLE_DEVICES"] = self.flags.cuda
        np.random.seed(self.flags.seed)
        self.supported_metrics = ["auroc", "macro_f1"]
        self.result = None
        # self.host_url = os.popen("ip -4 route show default").read().split()[2]      # Get the host machine's internal IP
        self.host_url = "127.0.0.1"

    def fit(self, X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray, y_test: np.ndarray, clf="ada",
            imp="gain", metric=None, args=None):
        """Train the selecting prediction pipeline selection system"""

        if self.data_folder_path is None:
            raise Exception("Please set the data folder path")

        # Set the parameters
        if args is not None:
            self.flags.metric = args.metric
            self.flags.dataset = args.dataset
        else:
            self.flags.metric = metric
            # TODO make the dataset transfer work
            self.flags.dataset = None
        # self.flags.metric = args.metric

        self.flags.clf = clf
        self.flags.data_names = ["X_train_autosmote.csv", "y_train_autosmote.csv", "X_test_autosmote.csv",
                                 "y_test_autosmote.csv"]
        #------------------
        # save x and y to data folder

        for data, name in zip([X_train, y_train, X_test, y_test], self.flags.data_names):
            data_path = os.path.join(self.data_folder_path, "interim", name)
            np.savetxt(data_path, data, delimiter=",")
            # df = pd.DataFrame(data)
            # df.to_csv(data_path, sep=",", header=False, index=False)

        # Start training
        client = docker.from_env()

        # Get the image build
        image_name = 'autosmote:version1.0'
        image_build_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "autosmote")
        try:
            client.images.get(image_name)
            logging.info('found image')
        except docker.errors.ImageNotFound:
            logging.info("Building AutoSMOTE 1.0 image")
            client.images.build(path=image_build_path, tag=image_name, nocache=True)

        # Create the container
        # volume1 = os.path.abspath("hybrids/autosmote")
        volume1 = image_build_path
        volume2 = os.path.abspath(self.data_folder_path)
        container_name = "autosmote-flask"

        logging.info('Creating AutoSMOTE container')
        try:
            container = client.containers.get(container_name)
            if container.status == "exited":
                container.start()
            logging.info('found container')
        except:
            container = client.containers.run(
                name=container_name,
                image=image_name,
                network_mode="host",
                # ports={"8080": 8080},
                volumes=[f'{volume1}:/code',
                         '/var/run/docker.sock:/var/run/docker.sock', # start container inside a container
                         f'{volume2}:/data'],
                entrypoint=["python3", "/code/app.py"],
                ipc_mode='host',
                detach=True,
            )
        timeout = 120
        stop_time = 3
        elapsed_time = 0
        logging.info('waiting container to be ready')
        while container.status != 'running' and elapsed_time < timeout:
            print(container.status)
            if container.status == 'exited':
                # TODO remove comment after project finish
                # container.remove(force=True, v=True)
                raise Exception("AutoSMOTE docker container build error")
            sleep(stop_time)
            elapsed_time += stop_time
            container = client.containers.get(container_name)
            continue

        # Wait the flask REST API to be ready
        sleep(5)

        # POST -- set parameters
        post_url = f'http://{self.host_url}:8080/set'
        headers =  {"Content-Type":"application/json"}
        response = requests.post(post_url, json.dumps(self.flags.__dict__), headers=headers)
        if response.status_code != 201:
            raise Exception("There is an error in setting AutoSMOTE parameters")

        # GET -- get result
        logging.info('Getting result from REST API')
        get_url = f'http://{self.host_url}:8080/results/' + self.flags.dataset

        # Delete container if error is encountered
        try:
            response_API = requests.get(get_url)
            self.result = response_API.json()['result']
        except Exception as e:
            container.remove(force=True, v=True)
            print(e)
            raise Exception("There is an error in getting AutoSMOTE result")


        # remove used container
        container.remove(force=True, v=True)

        # remove unuseful data files
        for name in self.flags.data_names:
            data_path = os.path.join(self.data_folder_path, "interim", name)
            if os.path.isfile(data_path):
                os.remove(data_path)

    def predict(self, X_test: np.ndarray = None, y_test: np.ndarray = None):
        return self.result


if __name__ == "__main__":
    logging.basicConfig(filename='cvd.log', level=logging.DEBUG,
                        format='%(asctime)s:%(levelname)s:%(message)s')

    warnings.filterwarnings("ignore")
    run_autosmote = RunAutoSmote()
    run_autosmote.fit(clf="mlp", imp="MIRACLE", metric="auroc", train_ratio=0.2)



import logging

import numpy as np
from sklearn.multiclass import OneVsRestClassifier
from sklearn import svm
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, ExtraTreesClassifier
from sklearn.metrics import precision_recall_fscore_support, roc_auc_score, classification_report, \
    average_precision_score

from xgboost import XGBClassifier
from sklearn.ensemble import GradientBoostingClassifier

clfs = {
    "lr": LogisticRegression(random_state=42, max_iter=500, solver='lbfgs'),
    "mlp": MLPClassifier(alpha=1, random_state=42, max_iter=1000),
    "ada": AdaBoostClassifier(random_state=42),
    "svm": svm.SVC(random_state=42, probability=True, kernel='linear'),
    # "SVM": svm.SVC(random_state=42, probability=True, kernel='poly'),
    # "rf": RandomForestClassifier(random_state=42),
    # "ensemble": XGBClassifier(learning_rate=1.0, max_depth=10, min_child_weight=15, n_estimators=100, n_jobs=1, subsample=0.8, verbosity=0),
    # "bst": GradientBoostingClassifier(random_state=42),
}

class CustomClassifier:
    def __init__(self, args):
        # def __init__(self, X: pd.DataFrame, Y: pd.DataFrame):
        self.args = args
        # self.f1 = None
        # self.precision = None
        # self.recall = None
        # self.w_precision = None
        # self.w_recall = None
        # self.auroc = None
        # self.auprc = None
        #
        # self.c_index = None
        self.clf = None
        self.clf_name = None

    def train(self, X_train: np.ndarray, Y_train: np.ndarray, clf=None):
        # Train classifier
        if clf in clfs.keys():
            self.clf = clfs[clf]
            self.clf_name = clf
        else:
            raise Exception("Model {} not defined in model.py".format(clf))

        self.clf.fit(X_train, Y_train)

    def predict(self, X_test: np.ndarray, Y_test: np.ndarray):

        if self.args.metric == "auroc":
            y_proba = self.clf.predict_proba(X_test)[:, 1]
            auroc = roc_auc_score(Y_test, y_proba)
            self.result = auroc
            return auroc
        elif self.args.metric == "macro_f1":
            y_pred = self.clf.predict(X_test)
            _, _, f1, _ = (
                precision_recall_fscore_support(Y_test, y_pred, average='macro'))
            self.result = f1
            return f1
        else:
            raise ValueError("Metric {} is not supported in {}".format(self.args.metric, self.clf_name))



import logging
import numpy as np
import pandas as pd
from sklearn.feature_selection import SelectKBest, chi2

class DataPreprocess:
    def __init__(self, data: pd.DataFrame, args):
        self.data = data
        self.X = None
        self.y = None
        self.header_X = None
        self.sel_header = None
        self.feature2drop = []
        self.args = args

    def split_data(self, target_name):
        header = list(self.data.columns.values)
        header.remove(target_name)

        self.X = self.data[header].copy()
        self.y = self.data[target_name].copy()
        logging.info("\t Total number of training data: {}, and CVD death {}".format(
            self.y.shape[0], np.sum(self.y)))

        # convert y to 0 and 1 if there is one 2 classes
        unique_classes = sorted(self.y.unique())
        if len(unique_classes) == 2:
            if set(unique_classes) != {0, 1}:
                # Create a mapping for the two classes
                mapping = {unique_classes[0]: 0, unique_classes[1]: 1}
                self.y = self.y.map(mapping)

        self.header_X = list(self.X.columns.values)

    def feature_selection(self, n_features=40):
        #     sel = VarianceThreshold(threshold=0.8*(1-0.8))
        sel = SelectKBest(chi2, k=n_features)
        sel.fit_transform(self.X, self.y)
        self.sel_header = self.X.columns.values[sel.get_support()]
        self.X = self.X[self.sel_header]

    def preprocess(self, args=None):
        # Split labels from data
        self.split_data(args.target_name) if args else self.split_data(self.args.target_name)

        return self.X, self.y

        # {{{ seperated from the main code
        # Sort data by "SEQN" for better performance in debugging
        # self.data_sort(["SEQN"])
        #
        # self.data_aggregation()
        #
        # self.delete_features()
        # }}}


    # def drop_columns(self, df):
    #     return df.drop([column for column in self.feature2drop if column in df.columns], axis=1)

    # def data_aggregation(self):
    #     """ Drop columns that are not useful and aggregation data to reduce dimension
    #     """
    #     # Generate the CVD death flag
    #     self.data = self.data.assign(Status=self.data['mortstat'] & (self.data['ucod_leading'] == 1))
    #
    #     # Diabetes Flag generation direction:
    #     #   1, Categorical variable of the following:
    #     #      0: Non-diabetes,
    #     #      1: Pre-diabetes,
    #     #      2: Diabetes
    #     #   2, Binary flag of diabetes
    #     def determine_value(fg, hba1c, dmdrug):
    #         """ For glycemicstatus feature
    #         """
    #         if fg < 100 and hba1c < 5.7:
    #             return 0
    #         elif 100 <= fg <= 125 or 5.7 <= hba1c <= 6.4:
    #             return 1
    #         elif fg >= 126 or hba1c >= 6.5 or dmdrug:
    #             return 2
    #         else:
    #             return np.nan
    #
    #     self.data['glycemicstatus'] = self.data.apply(
    #         lambda x: determine_value(x['fastglucose_ModP'], x['LBXGH'], x['DMdrug']),
    #         axis=1)
    #
    #     def combine_dm(self_report, exam, binary=False):
    #         """ For dm_flag feature
    #         """
    #         if not binary:
    #             # if the desire output is categorical
    #             if self_report == 1 or exam == 2:
    #                 return 2
    #             elif exam == 1:
    #                 return 1
    #             elif exam == 0:
    #                 return 0
    #             elif np.isnan(exam):
    #                 return np.nan
    #         else:
    #             # if the desire output is binary
    #             if self_report == 1 or exam == 2:
    #                 return 1
    #             elif self_report == 0 or exam == 0:
    #                 return 0
    #             else:
    #                 return np.nan
    #
    #     # TODO very next thing to do
    #     if self.args.aggregation == "categorical":
    #         self.data['dm_flag'] = self.data.apply(lambda x: combine_dm(x['DMdiagnosishistory'], x['glycemicstatus']),
    #                                                axis=1)
    #     elif self.args.aggregation == "binary":
    #         self.data['dm_flag'] = self.data.apply(
    #             lambda x: combine_dm(x['DMdiagnosishistory'], x['glycemicstatus'], binary=True),
    #             axis=1)
    #     self.feature2drop.append('glycemicstatus')

    # def delete_features(self):
    #
    #     # Delete participant IDs
    #     self.feature2drop.append('SEQN')
    #
    #     # Delete because duplication with "educationfor20"
    #     self.feature2drop.append('DMDEDUC2')
    #
    #     # Not useful as 'LBXGH' is already used in generating 'dm_flag'
    #     self.feature2drop.append('HBA1c_le65')
    #
    #     # Delete because of data aggregation for 'dm_flag'
    #     self.feature2drop.append('fastglucose_ModP')
    #     self.feature2drop.append('LBXGH')
    #     self.feature2drop.append('DMdrug')
    #
    #     # Delete unused prediction direction
    #     self.feature2drop.append('mortstat')
    #     self.feature2drop.append('ucod_leading')
    #     self.feature2drop.append('diabetes')
    #     self.feature2drop.append('hyperten')
    #
    #     # TODO check if need to keep with Dr Leung -- result: delete
    #     self.feature2drop.append('LBXSGL')
    #     self.feature2drop.append("glucose_nonfast_le200")
    #
    #     self.data = self.drop_columns(self.data)

    # def only_positive_cases(self):
    #     self.data = self.data.drop(self.data[self.data['cvddeath'] == 0].index)

    # def data_sort(self, value):
    #     self.data.sort_values(by=value, inplace=True)

